-- Model: Agents - Schelling's segregation model 
--
-- Author: Ondrej and Linda
--
-- Lesson: Agent based models in TerraME.
-- 		   Before add agents to an environment, add all cellular space the will need.
--		   Model MUST have the following structure
--				1) Define space
--				2) Define behavior (agents or societies or groups)
--				3) Define time
--				4) Define environments (adding (1) cellular spaces, then (2) agents, then (3) timers )
--				5) Run simulations 
--

--------------------------------------------------------------
-- PARAMETER 
--------------------------------------------------------------

--space params
INITCELLTEMP = 10
EVAPORATIONRATE = 0.04

-- cells params

-- society params
SPACE_DIMENSION = 100
SOCIETY_SIZE	= 50	

-- agents params
MINIDEALTEMP = 45
MAXIDEALTEMP = 60
MINOUTPUTHEAT = 10
MAXOUTPUTHEAT = 20
DIFFUSIONRATE = 0.9
RANDOMMOVECHANCE = 0.5

-- legend params
MINTEMP = 0
MAXTEMP = 150
HASAGENT = 1
-- simulationn temporal extent
FINAL_TIME = 100


--------------------------------------------------------------
-- GLOBAL 
--------------------------------------------------------------

-- random number generator
rand = Random()

--------------------------------------------------------------
-- MODEL 
--------------------------------------------------------------

-------------
--# SPACE #
-------------

-- This class behaves like a factory of cells. It creates cellular space, neighbourhood and assigns 
-- initial properties to the cells. Method evaporate() is repsonsible for constatnt heat loss of all cells 
CellsFactory = {
  cs = nil,
	init = function(self)
		self.cs = CellularSpace {
			xdim = SPACE_DIMENSION
		}
		self:doNeighbourhood()
		self:setCellProps()
		return self.cs
	end,
	doNeighbourhood = function(self)
		self.cs:createNeighborhood {
 			strategy = "moore",
			self = false
		}
	end,
	setCellProps = function(self)
		forEachCell(self.cs, function (cell)
			cell.temp = INITCELLTEMP
	  	cell.hasAgent = 0
	  	cell.color = 0
		end)
	end,
	evaporate = function(self)
		forEachCell(self.cs, function (cell)
			cell.temp = cell.temp*(1-EVAPORATIONRATE)
		end)
	end,
	scaleColor = function(self)
		forEachCell(self.cs, function (cell)
	  	cell.color = (cell.temp > MAXTEMP) and MAXTEMP or cell.temp
		end)
	end
}
cs = CellsFactory:init()


----------------
--# AGENT #
----------------
-- This is an agent prototype. It consist of setters of properties and core methods
AgentPrototype = Agent {
	-- unhappiness = 1,
	-- constructor
	init = function (self)
			self.color = 1
			self.idealTemp = 0
			self.outputHeat = 1
			self.unhappiness = 1
			self:setIdealTemp()
			self:setOutputHeat()
	end,
	-- executed in each iteration
	execute = function (self)
			self:diffuse()
		  self:step()
			self:heatCell()
	end,
	setIdealTemp = function(self)
		self.idealTemp =  MINIDEALTEMP + rand:number(MAXIDEALTEMP - MINIDEALTEMP)
	end,
	setOutputHeat = function(self)
		self.outputHeat =  MINOUTPUTHEAT + rand:number(MAXOUTPUTHEAT - MINOUTPUTHEAT)
	end,
	setUnhappiness = function(self)
		self.unhappiness = math.abs(self.idealTemp - self:getCell().temp)
	end,
	-- diffuse defined amount to neighbours
	diffuse = function(self)
		local cell = self:getCell()
		local amount = (cell.temp*DIFFUSIONRATE)/8
		forEachNeighbor(cell, function(c, n) 
			n.temp = n.temp + amount
		end)
	end,
	-- choose best neighbour according to unhappiness, move there
	step = function(self)
		self:setUnhappiness()
		if(self.unhappiness > 0) then
			local	bestCell = nil
			-- do step only if probability is reached
			if(rand:number(0,1) < RANDOMMOVECHANCE) then
				bestCell = self:decideBestCell()		--decide which cell is the best
			else
				bestCell = self:getCell():getNeighborhood():sample()
			end
			if(bestCell ~= nil) then
				self:moveTo(bestCell)
			end
		end
		-- heat cell
	end,
	-- decide which cell is best for me according to my unhappiness type (cold/hot)
	-- choose the cell with maximum property (hottest/coldest)
	decideBestCell = function(self)
		local currCell = self:getCell()
		local diff = 0
		local bestCell = currCell;
		-- print("My UNHAPPINESS is: "..currCell:getX()..self.unhappiness)
		-- print("My position is: ["..currCell:getX()..", "..currCell:getY().."]")
		-- print("My temp is:"..currCell.temp)
		-- print("My ideal-temp is:"..self.idealTemp)
		if  currCell.temp < self.idealTemp then		-- if i am cold
			--choose max temp
			diff = currCell.temp
			forEachNeighbor(currCell, function(c, n) 

				-- print("I am cold, but this cell is: "..n.temp)

				if n.temp > diff and n.hasAgent == 0 then
					diff = n.temp
					bestCell = n
				end
			end)
			-- print("this is my solution hottest: "..bestCell.temp)
		else -- if i am hot
			-- choose minimum temp
			diff = currCell.temp

			forEachNeighbor(currCell, function(c, n) 

				-- print("I am hot, but this cell is: "..n.temp)

				if n.temp < diff and n.hasAgent == 0 then
					diff = n.temp
					bestCell = n
				end
			end)
			-- print("this is my solution coldest: "..bestCell.temp)
		end
		-- print("--------------------------------------")
		return bestCell
	end,
	-- actually move to new cell
	moveTo = function(self, bestCell)
		if bestCell then
			self:getCell().hasAgent = 0
			self:move(bestCell)		-- move cell
			bestCell.hasAgent = 1
		end
	end,
	-- increase a temp of current cell
	heatCell = function(self)
		self:getCell().temp = self:getCell().temp + self.outputHeat		-- heat up my cell
	end
}

-- defines society
soc = Society {
	instance = AgentPrototype, 
	quantity = SOCIETY_SIZE,
}

-----------
--# TIME #
-----------
-- sequence of commands done by timers. 
t = Timer {
	Event {time = 1, period = 1, action = function(self)
		print("Time: "..self:getTime())
	end },
	--  1) diffuse
	Event {time = 1, period = 1, action = function()
		env:diffuseFromAllAgents()
		CellsFactory.cs:synchronize()
		CellsFactory.cs:notify()
	end},
	--	2) evaporate, scaleColor
	Event {time = 1, period = 1, action = function()
		CellsFactory:evaporate()
		CellsFactory:scaleColor()
		CellsFactory.cs:synchronize()
		CellsFactory.cs:notify()
	end},
	--	3) move agents
	Event {time = 1, period = 1, action = soc}
}

-------------------------------------------
--# ENVIRONMENT ( SCALE or VIRTUAL WORLD) #
-------------------------------------------
-- defines environment. It places agents to cells where max 1 agent/cell is allowed. 
-- sets agents unhappiness and cells property hasAgent. Calls diffuse() for each cells.
env = Environment { 
  cs = cs, 
  soc = soc, 
  t = t,
  init = function(self)
  	self:doPlacement()
		self:assignAgents()
	end,
	doPlacement = function(self)
		self:createPlacement {
			strategy = "random",
			max = 1 
		}
	end,
	assignAgents = function(self)
		forEachAgent (self.soc, function (agent)
				agent:setUnhappiness()
				print(agent.unhappiness)
				agent:getCell().hasAgent = 1
		end)
	end,
	diffuseFromAllAgents = function(self)
		forEachAgent (self.soc, function (agent)
				agent:diffuse()
		end)
	end
}
env:init()
--------------------------------------------------------------
-- SIMULATION
--------------------------------------------------------------
-- graphical simulation
Simulation = {
	init = function(self)
		self.agentsLegend = nil
		self.cellsLegend = nil
		self.map = nil
		self.socView = nil
		self:setCellsLegend()
		self:setAgentsLegend()
		self:setMapObs()
		self:setSocObs()
	end,
	setAgentsLegend = function(self)
	  self.agentsLegend = Legend {
			grouping = "uniquevalue",
			colorBar = {
        {color = "green"},
        { color = "yellow"}
    },
		symbol = "*",
		font = "Times",
		fontSize = 10
		}
	end,
	setCellsLegend = function(self)
		self.cellsLegend = Legend {
			grouping = "equalsteps",
			slices = 150,
			colorBar = {
		    {value = MINTEMP, color = "black"} ,
		    {value = MAXTEMP, color = "red"}
		  }
		}
	end,
	setMapObs = function(self)
		self.map = Observer {
			type = "map",
			subject = cs,
			attributes = {"color"},
			legends = {self.cellsLegend}
		}
	end,
	setSocObs = function(self)
		self.socView = Observer {
			type = "map",
			subject = soc,
			observer = self.map,
			attributes = {"color"},
			legends = {self.agentsLegend}
		}
	end
}

-- EXECUTE THE ENVIRONMENT
Simulation:init()
env:execute(FINAL_TIME)